import bcrypt
import hashlib
import six


WORK_FACTOR = 12  # 12=bcrypt default


HASH_FUNCTIONS = {
    'sha256': lambda plain, hashed: _encode(hashlib.sha256(plain).hexdigest()),
}


def check(plain, hashed):
    algorithm, sep, hashed_value = hashed.partition(':')
    if algorithm.lower() in HASH_FUNCTIONS:
        hash_fn = HASH_FUNCTIONS[algorithm.lower()]
        hashed = hashed_value
    else:
        hash_fn = bcrypt.hashpw
    plain = _encode(plain)
    hashed = _encode(hashed)
    return _constant_time_equal(hash_fn(plain, hashed), hashed)


def hash(plain):
    plain = _encode(plain)
    return bcrypt.hashpw(plain, bcrypt.gensalt(WORK_FACTOR)).decode('ascii')


def _constant_time_equal(a, b):
    # adapted from <http://throwingfire.com/storing-passwords-securely/>,
    # see also <http://codahale.com/a-lesson-in-timing-attacks/>
    if len(a) != len(b):
        return False

    result = 0
    if six.PY3:
        for x, y in zip(a, b):
            result |= x ^ y
    else:
        for x, y in zip(a, b):
            result |= ord(x) ^ ord(y)
    return result == 0


def _encode(text):
    if isinstance(text, six.text_type):
        text = text.encode('utf8')
    return text



import hashlib
import six
import unittest


class PasswordHashingTest(unittest.TestCase):

    def setUp(self):
        import gocept.loginuser.password
        import gocept.testing.patch
        self.patch = gocept.testing.patch.Patches()
        self.patch.set(gocept.loginuser.password, 'WORK_FACTOR', 4)

    def tearDown(self):
        self.patch.reset()

    def test_verifying_passwords(self):
        import gocept.loginuser.password
        hashed = gocept.loginuser.password.hash('mypassword')
        self.assertTrue(gocept.loginuser.password.check(
            'mypassword', hashed))
        self.assertFalse(gocept.loginuser.password.check(
            'invalid', hashed))

    def test_hash_accepts_and_returns_unicode(self):
        import gocept.loginuser.password
        hashed = gocept.loginuser.password.hash(u'mypassword')
        self.assertIsInstance(hashed, six.text_type)

    def test_sha256_password_is_recognised(self):
        import gocept.loginuser.password
        hashed = hashlib.sha256(b'asdf').hexdigest()
        self.assertTrue(gocept.loginuser.password.check(
            'asdf', 'sha256:' + hashed))

    def test_password_check_defaults_to_bcrypt(self):
        import gocept.loginuser.password
        hashed = gocept.loginuser.password.hash(u'asdf')
        self.assertTrue(gocept.loginuser.password.check(
            'asdf', hashed))



# Copyright (c) 2015 gocept gmbh & co. kg
# See also LICENSE.txt

# This should be only one line. If it must be multi-line, indent the second
# line onwards to keep the PKG-INFO file format intact.
"""Sqlalchemy user object and password management.
"""

from setuptools import setup, find_packages
import glob
import os.path


def project_path(*names):
    return os.path.join(os.path.dirname(__file__), *names)


setup(
    name='gocept.loginuser',
    version='1.2',

    install_requires=[
        'setuptools',
        'sqlalchemy',
        'bcrypt',
        'six',
    ],

    extras_require={
        'test': [
            'gocept.testing',
        ],
    },

    entry_points={
        'console_scripts': [
            # 'binary-name = gocept.loginuser.module:function'
        ],
    },

    author='gocept <mail@gocept.com>',
    author_email='mail@gocept.com',
    license='ZPL 2.1',
    url='https://bitbucket.org/gocept/gocept.loginuser/',

    keywords='Sqlalchemy user password login',
    classifiers="""\
Development Status :: 5 - Production/Stable
Intended Audience :: Developers
License :: OSI Approved
Natural Language :: English
Topic :: Database
Topic :: Software Development
Operating System :: OS Independent
License :: OSI Approved :: Zope Public License
Programming Language :: Python :: Implementation
Programming Language :: Python :: Implementation :: CPython
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 2.6
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
"""[:-1].split('\n'),
    description=__doc__.strip(),
    long_description='\n\n'.join(open(project_path(name)).read() for name in (
        'README.txt',
        'HACKING.txt',
        'CHANGES.txt',
    )),

    namespace_packages=['gocept'],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    data_files=[('', glob.glob(project_path('*.txt')))],
    zip_safe=False,
)



# encoding=utf-8
from sqlalchemy import Column, Integer, String, DateTime
import gocept.loginuser.password


class InvalidLockReasonError(ValueError):

    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return "%s is not a valid lock reason." % self.reason


class User(object):

    LOCK_STATUS_MAXLOGINS = 'Zu viele fehlerhafte Login-Versuche'
    LOCK_STATUS_ADMIN = 'Gesperrt durch Administrator'

    username = Column(String(254), unique=True)
    encrypted_password = Column('password', String(80))

    failed_logins = Column(Integer, nullable=False, server_default='0')
    is_locked = Column(String(50), default=u'')
    last_login = Column(DateTime, nullable=True)

    @classmethod
    def create(cls, **kw):
        username = kw.get('username')
        if username is not None:
            kw['username'] = username.lower()
        return super(User, cls).create(**kw)

    @classmethod
    def by_username(cls, username):
        """Get a user by its unique username or `None`."""
        return cls.query().filter_by(username=username.lower()).first()

    @property
    def password(self):
        return self.encrypted_password

    @password.setter
    def password(self, value):
        self.encrypted_password = self.hash_password(value)

    def hash_password(self, value):
        return gocept.loginuser.password.hash(value)

    def check_password(self, plain):
        return gocept.loginuser.password.check(plain, self.password)

    @property
    def authenticated(self):
        return self.username is not None

    def lock(self, reason=None):
        """Lock a user because of `reason`.

           Reason can be:
               "MAXLOGINS": Too many unsuccessful login retries.
               "ADMIN": Locked by admin via admin ui.
        """
        reason_text = getattr(self, 'LOCK_STATUS_%s' % reason, None)
        if reason_text is None:
            raise InvalidLockReasonError(reason)
        self.is_locked = reason_text

    def unlock(self):
        """Unlock a user."""
        self.is_locked = None

    def __str__(self):
        return '<User %s>' % self.username
